########## simulation runs (from original apm paper but now comparing Augmented Lagrangian and ConstOptim)
########## 

################################ SIMULATIONS

# libraries required 

# plyr is for determining unique covariates (could also use melt())
library(plyr)
library(profvis)
# numDeriv is for numerical derivatives for hessian/score
library(numDeriv)
# Matrix give a rankMatrix function for the rank of a matrix
library(Matrix)
library(alabama)
library(rapportools)
library(Rcpp)

trace <- FALSE
case <- 'A'

######################
setTestCase <- function(incase = "A", intrace = FALSE, basedir) {
  
    if(!incase %in% list("A", "E", "F", "G")){
      print("Invalid case")
      stop()
    }
    
    case <<- incase
    
    if(incase == "A"){
          source(paste(basedir, "/source/R/lcpmMinusloglik0A.R",sep=""),keep.source=TRUE)
     } else if(incase == "E"){
             library(dplyr)        
             source(paste(basedir, "/source/R/lcpmMinusloglik0A.R",sep=""),keep.source=TRUE)
      } else if (incase == "F"){
              library(dplyr)
              library(Rcpp)
              rm('lcpmMinusloglik0A.R')
              cppFile1 <- paste(basedir,"/source/cpp/loglik101.cpp",sep="")
              #cppFile2 <- paste(basedir,"/source/cpp/loglik100.cpp",sep="")
              #print(cppFile2)
              #source(paste(basedir,"/source/R/lcpmMinusloglik0B.R",sep=""),keep.source=TRUE)
              sourceCpp(cppFile1)
              #sourceCpp(cppFile2)
      } else if (incase == "G") {
                library(dplyr)
                library(Rcpp)
                #rm(list = c('lcpmMinusloglik0A.R','loglik100.cpp'))
                library(loglikrad98)
                #source(paste(basedir, "/source/R/lcpmMinusloglik0B.R",sep=""),keep.source=TRUE)
      } 
        
       
      
    
#    print(paste('case: ', case, ' incase ', incase))    
    
    if (is.boolean(intrace)) {
      trace <<- intrace
    } else {
        print("Invalid TRACE INDICATOR")
        stop()
    }
  
  }  
  

generateDataFormula <- function(incase = 'A', seedi, r, n, nc) {
  
    # generate Covariates
    seedir<-seedi+r
    set.seed(seedir)
    x1<-matrix(rbinom(nc*n,size=1,prob=0.5),byrow=TRUE,ncol=nc,nrow=n)
    colnames(x1) <- paste0("x", 1:nc)

    # generate response variables
    alpha1<- -1-0.1*nc/2 
    alpha2<- -0.1*nc/2
    beta1<- c(rep(-0.1,nc/2),rep(0.1,nc/2))
    p1<-as.vector(exp(alpha1+x1%*%beta1))
    p2<-as.vector(exp(alpha2+x1%*%beta1)-exp(alpha1+x1%*%beta1))
    p3<-as.vector( 1-exp(alpha2+x1%*%beta1))
    y<-rep(NA,n)
    for(i in 1:n){
      y[i]<-sample(c(1:3),1,replace=TRUE,prob=c(p1[i],p2[i],p3[i]))
    }
    
    data1<-cbind(y,x1)
    mydata<-as.data.frame(data1)
    # CHANGE as based on problem
    # number of covariates (not intercepts) in model
    #print(paste('mydata ', nrow(mydata), 'columns: ', ncol(mydata)))
    model.columns <- colnames(mydata)
    xcols <- paste(model.columns[2:nc],collapse=" + ")
    fmla <- as.formula(paste('y ~', xcols))
    

    model.columns <- colnames(mydata)
    xcols <- paste(model.columns[1:nc+1],collapse=" + ")
    fmla <- as.formula(paste('y ~', xcols))
    
    return(list(as.data.frame(mydata), as.formula(fmla)))

}

generateSumdat <- function(incase = 'A', fmla, data) {

  #browser()
  if (incase == 'A') {  
  
    Y<-model.frame(fmla, data)[, 1]
  
    X<- model.matrix(fmla, data)
    mydata<-na.omit(as.data.frame(cbind(Y,X[,-1])))
    colnames(mydata)<-c("y",colnames(X)[-1])
    sumdat<-plyr::count(mydata)

  } else if (incase %in% list('E', 'F', 'G')) {
    
    library(dplyr)
    #browser()
    mydata <- model.frame(fmla, data,na.action = na.omit)
    #print(mydata)
    colnames(mydata)[1] = 'y'
    #print(colnames(mydata))
    x.col.count <- length(colnames(mydata)[colnames(mydata) !='y'])
    #print(x.col.count)
    mydata$y <- as.numeric(as.factor(mydata$y))
    out <- names(Filter(is.factor,mydata))[]
    if (length(out) > 0) {
      colnames <- out[!(out %in% c('y'))]
      mydata <- fastDummies::dummy_cols(mydata, select_columns=colnames,remove_selected_columns=TRUE, remove_first_dummy=TRUE)
    } 
    group <- colnames(mydata)
    #print(group)
    sumdat = data.frame(mydata %>% group_by_at(group) %>% dplyr::summarise(freq = n()))
    #print(sumdat[1:5,])
    
  
  } 
  return (sumdat)
  
}  

checkElapsedTime <-function(stime) {
  
  ctime = Sys.time()    
  elapsed = round(as.numeric(difftime(time1 = ctime, time2 = stime, units = "hours")), 3)
  return (elapsed > 5) 
}

############################################################################################################

#             EXAMPLES (Boundary)

############################################################################################################
executeTests <- function(case, replicates, sampleSize, covariates, seed, basedir) {
  rlog::log_info(paste("execute Tests",case, replicates, sampleSize, covariates, seed, basedir, sep=" "))
  #print(paste(case, replicates, sampleSize, covariates, seed, basedir))
  currentDate <- format(Sys.Date(), "%m-%d-%y")
  #rpfilename <- paste(basedir, '/output/profiles/P0', covariates, case,'-timings-', currentDate, '-T',sampleSize, '.prof',sep='' )
  #Rprof(filename=rpfilename,memory.profiling=TRUE,line.profiling=TRUE)
  setTestCase(incase = case, intrace = FALSE, basedir)

  # errors in Auglag and nelder mead SE computation (full run no boundaries?)
  N<- replicates # number of replicates
  n<- sampleSize # sample size
  # algorithm requires nc to be even number
  nc <- covariates # number of covariates 
  outmatrix<-matrix(rep(NA,N*26),nrow=N , ncol= 26 )
  colnames(outmatrix)<-c("seed","loglik_neldmead","boundary0_neldmead","boundary1_neldmead",  
                         "loglik_auglag","boundary0_auglag","boundary1_auglag","conv_neldmead",
                         "conv_auglag","t_user_neldmead", "t_sys_neldmead", "t_elapsed_neldmead",
                         "t_1_neldmead", "t_2_neldmead", "t_user_auglag", "t_sys_auglag",
                         "t_elapsed_auglag", "t_1_auglag", "t_2_auglag", "t_user_sumdat", 
                         "t_sys_sumdat", "t_elapsed_sumdat", "t_1_sumdat", "t_2_sumdat",
                         "alld_neldmead","alld_auglag")
  
  outtimings<-matrix(rep(NA,N*20),nrow=N , ncol= 20 )
  colnames(outtimings)<-c('case', 'currentDate', 'execution', 'n', 'covariates',"t_user_neldmead", 
                          "t_sys_neldmead", "t_elapsed_neldmead", "t_1_neldmead", "t_2_neldmead", 
                          "t_user_auglag", "t_sys_auglag", "t_elapsed_auglag", "t_1_auglag", 
                          "t_2_auglag", "t_user_sumdat", "t_sys_sumdat", "t_elapsed_sumdat", 
                          "t_1_sumdat", "t_2_sumdat")
  seedi<-seed
  stime = Sys.time()
  for(r in 1:N){
  
    if (checkElapsedTime(stime)) { 
        print('Time Elapsed Before Completion') 
        cat('executions within time =', r, '\n')
        break }
    
    seedir<-seedi
    set.seed(seedir)
    #browser()
    outdata <- generateDataFormula(case, seedi, r, n, nc)
    mydata  <- outdata[[1]]
    max(mydata$y)
    fmla <- outdata[[2]]
  
    # CHANGE as based on problem
    # number of covariates (not intercepts) in model
    p = nc
  
    #####################
    # identify the counts of unique strata/groups
    #sumdat<-count(mydata,colnames(mydata))

    try(time_sumdat<-system.time(sumdat <- generateSumdat(case, fmla, mydata)))
    #mydata
    ### construct a constraints matrix
  
    miny=1
    maxy=max(sumdat$y)
  
    ncuts=length(unique(mydata$y))-1
  
    Xy<-sumdat
  
    X1<-Xy[Xy$y==miny,]
    XJ<-Xy[Xy$y==maxy,]
    Xj<-Xy[Xy$y!=miny & Xy$y!=maxy,]
  
  
    # first ordinal value
    l1<-length(X1$y)
    Xa1<-cbind(matrix(c(rep(1,l1),rep(0,l1*(maxy-2))),ncol=maxy-1,byrow=FALSE),X1[,-1])
    Xa1<-as.matrix(Xa1)
    # last ordinal value
    lJ<-length(XJ$y)
    XaJ<-cbind(matrix(c(rep(0,lJ*(maxy-2)),rep(1,lJ)),ncol=maxy-1,byrow=FALSE),XJ[,-1])
    XaJ<-as.matrix(XaJ)
  
    # all other ordinal values
    lj<-length(Xj$y)
    Xaj1<-cbind(matrix(0,ncol=maxy-1,nrow=lj),Xj)
    Xaj1[cbind(seq(1,lj,1),Xaj1$y)]<-1
    Xaj1<-as.matrix(Xaj1[,-maxy])
  
    Xaj2<-cbind(matrix(0,ncol=maxy-1,nrow=lj),Xj)
    Xaj2[cbind(seq(1,lj,1),Xaj2$y-1)]<-1
    Xaj2<-as.matrix(Xaj2[,-maxy])
    
  
    extraconst=cbind(diag(rep(1,ncuts)),matrix(0,ncol=p+1,nrow=ncuts))
    Xa<-rbind(Xa1,XaJ,Xaj1,Xaj2,-(Xaj1-Xaj2),extraconst)
  
    #  Xa<-rbind(Xa1,XaJ,Xaj1,Xaj2,-(Xaj1-Xaj2))
    #  Xas<- (Xaj1-Xaj2)
  
    #drop last column which are counts
    Xa.f<<-as.matrix(Xa[,-ncol(Xa)])
  
    #Xas.f<<-as.matrix(Xas[,-ncol(Xas)])
    ## construct data matrix
  
    alphastarts<-c(-1,rep(-0.1,ncuts-1))
    startval<-c(rev(cumsum(alphastarts)),rep(0,p))
  
  
    {control.list=list(maxit=100000,abstol=1e-12,reltol=1e-12)}
    {eps.outer=1e-10}
  
    # constrained optimization
    #control.outer <- list(method="Nelder-Mead", NMinit=TRUE, kkt2.check=TRUE, eps=eps.outer, itmax=100000, i.scale=1, trace=TRUE, lam0=10, sig0=100)
  
  
    #try(time_neldmead<-system.time(constr_optim_neldmead<-constrOptim(startval,minusloglik.cir, Xa1=Xa1,XaJ=XaJ,Xaj1=Xaj1,Xaj2=Xaj2,ui=rbind(Xa.f,-Xa.f,Xas.f), ci=c(rep(0,length(Xa.f[,1])),rep(-1,length(Xa.f[,1])),rep(0,length(Xas.f[,1]))), method="Nelder-Mead",control=control.list,outer.eps=eps.outer,outer.iterations = 100000) ))
    rlog::log_info(paste("calling constrOptim",r, sep=" "))
    try(time_neldmead<-system.time(constr_optim_neldmead<-constrOptim(startval,minusloglikppm, Xa1=Xa1,XaJ=XaJ,Xaj1=Xaj1,Xaj2=Xaj2,Xaf=Xa.f,ui=-Xa.f, ci=rep(0,length(Xa.f[,1])), method="Nelder-Mead",control=control.list,outer.eps=eps.outer,outer.iterations = 100000)))
  
  
    # methods - with gradient BFGS, CG
    rlog::log_info(paste("calling auglag",r, sep=" "))
    time_auglag<-system.time(constr_optim_auglag<-auglag(par=startval, fn=minusloglikppm,  hin=h, Xa1=Xa1,XaJ=XaJ,Xaj1=Xaj1,Xaj2=Xaj2,Xaf=-Xa.f,
                                                           control.optim = control.list, control.outer=list(eps=1e-12,itmax=100000,trace=FALSE)))
    
    if (inherits(time_auglag, "try-error")) {
        error=function(e)
        RLOG::LOG_ERROR(" An Auglag Error has occurred: ", e$message)
        constr_optim_auglag <- NULL
        next
    }
    ################## NeldMead output
    mle_neldmead <- rep(NA,nc)
    loglik.max_neldmead <- NA 
    fitted.val_neldmead <- NA
    Boundary0_neldmead<-NA
    Boundary1_neldmead<-NA
  
    mle_neldmead <- constr_optim_neldmead$par
    loglik.max_neldmead <- -constr_optim_neldmead$value 
    fitted.val_neldmead <- exp(rbind(Xa1,Xaj1)[,-ncol(Xa)]%*%mle_neldmead)
    Boundary0_neldmead<-ifelse(any(fitted.val_neldmead < 1e-5),1,0)
    Boundary1_neldmead<-ifelse(any(fitted.val_neldmead > 0.99999),1,0)
    conv_neldmead<-constr_optim_neldmead$convergence
    # numerical derivative for approximating Hessian 
    #try(NegHess <- hessian(minusloglik.cir,constr_optim_neldmead$par,Xa1=Xa1,XaJ=XaJ,Xaj1=Xaj1,Xaj2=Xaj2))
    #SE_constrOptim_neldmead<-rep(NA,nc)
    #try(SE_constrOptim_neldmead<-sqrt(diag(solve(NegHess))))
  
    ################### AUGLAG
    mle_auglag <- rep(NA,nc)
    loglik.max_auglag <- NA 
    fitted.val_auglag <- NA
    Boundary0_auglag<-NA
    Boundary1_auglag<-NA
    mle_auglag <- constr_optim_auglag$par
    #browser()
    
    loglik.max_auglag <- -constr_optim_auglag$value 
    fitted.val_auglag <- exp(rbind(Xa1,Xaj1)[,-ncol(Xa)]%*%mle_auglag)
    Boundary0_auglag<-ifelse(any(fitted.val_auglag < 1e-5),1,0)
    Boundary1_auglag<-ifelse(any(fitted.val_auglag > 0.99999),1,0)
    #SE_auglag<-rep(NA,nc)
    #try(SE_auglag<-sqrt(diag(solve(constr_optim_auglag$hessian))))
    
    conv_auglag<-constr_optim_auglag$convergence
    #diffnum<-c(abs(mle_neldmead-mle_auglag),abs(loglik.max_neldmead-loglik.max_auglag))
    
    alld_neldmead <- abs(loglik.max_neldmead - max(loglik.max_neldmead, loglik.max_auglag))
    alld_auglag <- abs(loglik.max_auglag - max(loglik.max_neldmead, loglik.max_auglag))
  #  print(r)
    options(digits=9)
    outmatrix[r,]<-c(seedir,loglik.max_neldmead,Boundary0_neldmead,Boundary1_neldmead,loglik.max_auglag,Boundary0_auglag,Boundary1_auglag,conv_neldmead,conv_auglag,time_neldmead,time_auglag,time_sumdat,alld_neldmead,alld_auglag)
    
    currentDate <- format(Sys.Date(), "%m-%d-%y")
    outtimings[r,]<-c(as.character(case), currentDate, as.integer(r), as.integer(n), as.integer(nc),
                      as.numeric(time_neldmead[1]),as.numeric(time_neldmead[2]), as.numeric(time_neldmead[3]), 
                      as.numeric(time_neldmead[4]), as.numeric(time_neldmead[5]),
                      as.numeric(time_auglag[1]),as.numeric(time_auglag[2]), as.numeric(time_auglag[3]), 
                      as.numeric(time_auglag[4]), as.numeric(time_auglag[5]),
                      as.numeric(time_sumdat[1]),as.numeric(time_sumdat[2]), as.numeric(time_sumdat[3]), 
                      as.numeric(time_sumdat[4]), as.numeric(time_sumdat[5]))
                      
  }

  currentDate <- format(Sys.Date(), "%m-%d-%y")
  
  pfilename <- paste(basedir,'/output/P0', nc, case,'-PPMcovars-', currentDate, '-T',n,'.out',sep='' )
  print(pfilename)
  write.table(outmatrix, file=pfilename,sep="\t",row.names=FALSE)

  tfilename <- paste(basedir, '/output/csv/P0', nc, case,'-timings-', currentDate, '-T',n, '.csv',sep='' )
  print(tfilename)
  write.table(outtimings, file=tfilename,sep=",",row.names=FALSE)


  ### metrics


  # ALLD

  mean(outmatrix[,20])
  mean(outmatrix[,21])
#  Rprof(filename=NULL)
#  summary(outmatrix)
}


