########## simulation runs - basic construction by G.Singh on Aug 16. Adapted from APM simulations presented at JSM2023.  
###############


################################ SIMULATIONS

# libraries required 

# plyr is for determining unique covariates (could also use melt())
library(dplyr)
library(plyr)
library(numDeriv)
library(Matrix)
#library(VGAM)
#library(haven)
library(alabama)
library(profvis)
library(rlog)
library(alabama)
library(rapportools)
library(Rcpp)

######################
trace <- FALSE
case <- 'A'
alpha1<- 0
alpha2<- 0
beta1<- 0
beta2<- 0


######################
setTestCase <- function(incase = "A", intrace = FALSE, basedir) {
  
  rlog::log_trace(paste("running setTestCase ", incase))
  if(!incase %in% list("A", "E", "F", "G")){
    print("Invalid case")
    stop()
  }
  
  case <<- incase
  
  if(incase == "A"){
    source(paste(basedir, "/source/R/lcpmMinusloglik0A.R",sep=""),keep.source=TRUE)
  } else if(incase == "E"){
    library(dplyr)        
    source(paste(basedir, "/source/R/lcpmMinusloglik0A.R",sep=""),keep.source=TRUE)
  } else if (incase == "F"){
    library(dplyr)
    library(Rcpp)
    #rm('lcpmMinusloglik0A.R')
    cppFile <- paste(basedir,"/source/cpp/loglik100.cpp",sep="")
    print(cppFile)
    source(paste(basedir,"/source/R/lcpmMinusloglik0B.R",sep=""),keep.source=TRUE)
    sourceCpp(cppFile)
  } else if (incase == "G") {
    library(dplyr)
    library(Rcpp)
    #rm(list = c('lcpmMinusloglik0A.R','loglik100.cpp'))
    library(loglikrad98, lib.loc="D:/ProgramFiles/anaconda3/envs/ordinal/Lib/R/library")
    source(paste(basedir, "/source/R/lcpmMinusloglik0B.R",sep=""),keep.source=TRUE)
  } 
  
  
  
  
  print(paste('case: ', case, ' incase ', incase))    
  
}  


generateDataFormula <- function(incase = 'A', seedi, n, nc) {
  
  rlog::log_trace(paste("running generateDataFormula ", incase, n, nc))
  # generate Covariates
  x1<-matrix(rbinom(nc*n,size=1,prob=0.5),byrow=TRUE,ncol=nc,nrow=n)
  colnames(x1) <- paste0("x", 1:nc)
  
  # generate response variables
  #setup how to simulate data (Scenarios can be setup with a switch statement)
  alpha1<<- -0.2
  alpha2<<- -0.1
  beta1<<- -1
  beta2<<- -0.5
  x1<-rbinom(n,size=1,prob=0.5)
  x2<-runif(n,1,5)
  p1<-as.vector(exp(alpha1+beta1*x1+beta2*x2))
  p2<-as.vector(exp(alpha2+beta1*x1+beta2*x2)-exp(alpha1+beta1*x1+beta2*x2))
  p3<- as.vector(1-exp(alpha2+beta1*x1+beta2*x2))
  y<-rep(NA,n)
  for(i in 1:n){
    y[i]<-sample(c(1:3),1,replace=TRUE,prob=c(p1[i],p2[i],p3[i]))
  }
  data1<-cbind(y,x1,x2)  
  mydata<-as.data.frame(data1)
  
  model.columns <- colnames(mydata)
  xcols <- paste(model.columns[2:3],collapse=" + ")
  rlog::log_trace(paste("creating formula ", xcols))
  cat('\n','xcols', xcols, 'model.columns', model.columns)
  fmla <- as.formula(paste('y ~', xcols))
  
  return(list(as.data.frame(mydata), as.formula(fmla)))
  
}

generateSumdat <- function(incase = 'A', fmla, data) {
  #browser()
  rlog::log_trace(paste("generating sumDat ", incase))
  if (incase == 'A') {  
    #cat('got to sumdat 1')
    Y<-model.frame(fmla, data)[, 1]
    
    X<- model.matrix(fmla, data)
    mydata<-na.omit(as.data.frame(cbind(Y,X[,-1])))
    colnames(mydata)<-c("y",colnames(X)[-1])
    #cat('columns', colnames(mydata))
    sumdat<-plyr::count(mydata)
    rlog::log_trace(paste("finishing sumDat", dim(sumdat)))
  } else if (incase %in% list('E', 'F', 'G')) {
    
    library(dplyr)
    mydata <- model.frame(fmla, data,na.action = na.omit)
    colnames(mydata)[1] = 'y'
    x.col.count <- length(colnames(mydata)[colnames(mydata) !='y'])
    mydata$y <- as.numeric(as.factor(mydata$y))
    out <- names(Filter(is.factor,mydata))[]
    #cat('out:',out)
    if (length(out) > 0) {
      colnames <- out[!(out %in% c('y'))]
      #cat('out2:',length(out))
      mydata <- fastDummies::dummy_cols(mydata, select_columns=colnames,remove_selected_columns=TRUE, remove_first_dummy=TRUE)
    } 
    group <- colnames(mydata)
    #cat('cols:',group)
    sumdat <- data.frame(mydata %>% group_by_at(group) %>% dplyr::summarise(freq = n()))
    #print(paste('first time ', object.size(sumdat),units="Mb"))
    rlog::log_trace(paste("finishing generating sumDat ", object.size(sumdat)))
    
  } 
  return (sumdat)
  
}  

checkElapsedTime <-function(stime) {
  
  ctime = Sys.time()    
  elapsed = round(as.numeric(difftime(time1 = ctime, time2 = stime, units = "hours")), 3)
  return (elapsed > 8) 
}



############################################################################################################

#             EXAMPLES (Boundary)

############################################################################################################
executeTests <- function(case, replicates, sampleSize, covariates, seed, basedir) {
  rlog::log_info(paste("starting execute Tests ", case, replicates, sampleSize, covariates, seed, basedir))
  currentDate <- format(Sys.Date(), "%m-%d-%y")
  #rpfilename <- paste(basedir, '/output/profiles/P0', covariates, case,'-timings-', currentDate, '-T',sampleSize, '.prof',sep='' )
  #Rprof(filename=rpfilename,memory.profiling=TRUE,line.profiling=TRUE)
  rlog::log_trace(paste("calling setTestCase ", basedir))
  setTestCase(incase = case, intrace = FALSE, basedir)
  
  # errors in Auglag and nelder mead SE computation (full run no boundaries?)
  
  # errors in Auglag and nelder mead SE computation (full run no boundaries?)
  N<- replicates  # N is the number of replicates (can reduce this if variability is low: 2500 - 5000)
  n<- sampleSize  # n is the sample size (change this to a vector {500, 5000, 50000, 500000, 5000000, ...})
  nc<- covariates 
  outmatrix<-matrix(rep(NA,N*42),nrow=N , ncol= 42 )
  colnames(outmatrix)<-c("seed","optima1_neldmead","optima2_neldmead","optimb1_neldmead","optimb2_neldmead",
                         "loglik_neldmead","boundary0_neldmead","boundary1_neldmead", "SE_a1_neldmead","SE_a2_neldmead","SE_b1_neldmead","SE_b2_neldmead",
                         "optima1_auglag","optima2_auglag","optimb1_auglag","optimb2_auglag",  
                         "loglik_auglag","boundary0_auglag","boundary1_auglag", "SE_a1_auglag","SE_a2_auglag","SE_b1_auglag","SE_b2_auglag",
                         "diff_a1", "diff_a2","diff_b1","diff_b2", "diff_loglik","conv_neldmead","conv_auglag","t_user_neldmead", "t_sys_neldmead", "t_elapsed_neldmead", "t_1_neldmead", "t_2_neldmead", 
                         "t_user_auglag", "t_sys_auglag", "t_elapsed_auglag", "t_1_auglag", "t_2_auglag","alld_neldmead","alld_auglag")
  
  
  seedi<- seed
  rlog::log_trace(paste("calling generate Data Formula ", case, n, nc))
  outdata <- generateDataFormula(case, seedi, n, nc)
  
  rlog::log_trace(paste("beginning execute loop ", N)) 
  
  stime = Sys.time()  
  for(r in 1:N){
    rlog::log_trace(paste("doing execute loop", r))
    
    
    if (checkElapsedTime(stime)) { 
      print('Time Elapsed Before Completion') 
      cat('\n','executions within time =', r, '\n')
      break }
    
    seedir<-seedi+r
    set.seed(seedir)
    mydata  <- outdata[[1]]
    #cat('columns: ', colnames(mydata))
    max(mydata$y)
    fmla <- outdata[[2]]  
    p = nc
  
    rlog::log_trace(paste("calling generate sumdat ", case, nrow(mydata)))
    try(time_sumdat<-system.time(sumdat <- generateSumdat(case, fmla, mydata)))
    
    #mydata
    ### construct a constraints matrix
    miny=1
    maxy=max(sumdat$y)
    
    ncuts=length(unique(mydata$y))-1
    
    Xy<-sumdat
    ncuts=length(unique(mydata$y))-1
  
    Xy<-sumdat
  
    X1<-Xy[Xy$y==miny,]
    XJ<-Xy[Xy$y==maxy,]
    Xj<-Xy[Xy$y!=miny & Xy$y!=maxy,]
  
  
  # first ordinal value
    l1<-length(X1$y)
    Xa1<-cbind(matrix(c(rep(1,l1),rep(0,l1*(maxy-2))),ncol=maxy-1,byrow=FALSE),X1[,-1])
    Xa1<-as.matrix(Xa1)
  # last ordinal value
    lJ<-length(XJ$y)
    XaJ<-cbind(matrix(c(rep(0,lJ*(maxy-2)),rep(1,lJ)),ncol=maxy-1,byrow=FALSE),XJ[,-1])
    XaJ<-as.matrix(XaJ)
  
  # all other ordinal values
    lj<-length(Xj$y)
    Xaj1<-cbind(matrix(0,ncol=maxy-1,nrow=lj),Xj)
    Xaj1[cbind(seq(1,lj,1),Xaj1$y)]<-1
    Xaj1<-as.matrix(Xaj1[,-maxy])
  
    Xaj2<-cbind(matrix(0,ncol=maxy-1,nrow=lj),Xj)
    Xaj2[cbind(seq(1,lj,1),Xaj2$y-1)]<-1
    Xaj2<-as.matrix(Xaj2[,-maxy])
    
    #extraconst=cbind(diag(rep(1,ncuts)),matrix(0,ncol=p+1,nrow=ncuts))
    cat('\n','ncuts:', ncuts, 'p:', p)
    extraconst=cbind(diag(rep(1,ncuts)),matrix(0,ncol=p+1,nrow=ncuts))
    #cat('Xa1:', dim(Xa1), 'XaJ:', dim(XaJ),'Xaj1:', dim(Xaj1), 'Xaj2:', dim(Xaj2), 'extraconst:', dim(extraconst))
    Xa<-rbind(Xa1,XaJ,Xaj1,Xaj2,-(Xaj1-Xaj2),extraconst)

    #drop last column which are counts
    Xa.f<<-as.matrix(Xa[,-ncol(Xa)])
  
    rlog::log_debug(paste("Xa matrices have been generated ", 
                          "Xa1 ", dim(Xa1), "XaJ ", dim(XaJ), "Xaj1 ", dim(Xaj1), "Xaj2 ", dim(Xaj2), "Xa.f ", dim(Xa.f)))
    #Xas.f<<-as.matrix(Xas[,-ncol(Xas)])
    ## construct data matrix
  
    alphastarts<-c(-1,rep(-0.1,ncuts-1))
    startval<-c(rev(cumsum(alphastarts)),rep(0,p))
  
  
    {control.list=list(maxit=100000,abstol=1e-12,reltol=1e-12)}
    {eps.outer=1e-10}
  
    # constrained optimization
    #control.outer <- list(method="Nelder-Mead", NMinit=TRUE, kkt2.check=TRUE, eps=eps.outer, itmax=100000, i.scale=1, trace=TRUE, lam0=10, sig0=100)


    #try(time_neldmead<-system.time(constr_optim_neldmead<-constrOptim(startval,minusloglik.cir, Xa1=Xa1,XaJ=XaJ,Xaj1=Xaj1,Xaj2=Xaj2,ui=rbind(Xa.f,-Xa.f,Xas.f), ci=c(rep(0,length(Xa.f[,1])),rep(-1,length(Xa.f[,1])),rep(0,length(Xas.f[,1]))), method="Nelder-Mead",control=control.list,outer.eps=eps.outer,outer.iterations = 100000) ))
  
    rlog::log_debug(paste("calling neldmead ", r))
    try(time_neldmead<-system.time(constr_optim_neldmead<-constrOptim(startval,minusloglikppm, Xa1=Xa1,XaJ=XaJ,Xaj1=Xaj1,Xaj2=Xaj2,Xaf=Xa.f,ui=-Xa.f, ci=rep(0,length(Xa.f[,1])), method="Nelder-Mead",control=control.list,outer.eps=eps.outer,outer.iterations = 100000)))

    #browser()
  # methods - with gradient BFGS, CG
    rlog::log_debug(paste("calling auglag ", r))
    try(time_auglag<-system.time(constr_optim_auglag<-auglag(par=startval, fn=minusloglikppm,  hin=h ,
                                                             Xa1=Xa1,XaJ=XaJ,Xaj1=Xaj1,Xaj2=Xaj2,Xaf=-Xa.f,
                                                           control.optim = control.list, control.outer=list(eps=1e-12,itmax=100000,trace=FALSE))))

    rlog::log_debug(paste("back from optimizers ", r))
    ################## NeldMead output
    mle_neldmead <- c(NA,NA,NA,NA)
    loglik.max_neldmead <- NA 
    fitted.val_neldmead <- NA
    Boundary0_neldmead<-NA
    Boundary1_neldmead<-NA
  
    mle_neldmead <- constr_optim_neldmead$par
    loglik.max_neldmead <- -constr_optim_neldmead$value 
    fitted.val_neldmead <- exp(rbind(Xa1,Xaj1)[,-ncol(Xa)]%*%mle_neldmead)
    Boundary0_neldmead<-ifelse(any(fitted.val_neldmead < 1e-5),1,0)
    Boundary1_neldmead<-ifelse(any(fitted.val_neldmead > 0.99999),1,0)
    conv_neldmead<-constr_optim_neldmead$convergence
  # numerical derivative for approximating Hessian 
    NegHess <- try(hessian(minusloglikppm,constr_optim_neldmead$par,Xa1=Xa1,XaJ=XaJ,Xaj1=Xaj1,Xaj2=Xaj2))
    if (inherits(NegHess,"try-error")) {
      error = function(e)
      rlog::log_error("A NegHess error occurred: ", e$message)
      NegHess <- NULL
    } else  {
      SE_constrOptim_neldmead<-c(NA,NA,NA,NA)
      SE_constrOptim_neldmead<-try(sqrt(diag(solve(NegHess))))
      if (inherits(SE_constrOptim_neldmead,"try-error")) {
        error = function(e)
          rlog::log_error("A SE error occurred: ", e$message)
        SE_constrOptim_neldmead <- NULL
      }
    }  
         
  
    ################### AUGLAG
    mle_auglag <- c(NA,NA,NA,NA)
    loglik.max_auglag <- NA 
    fitted.val_auglag <- NA
    Boundary0_auglag<-NA
    Boundary1_auglag<-NA
  
    mle_auglag <- constr_optim_auglag$par
    loglik.max_auglag <- -constr_optim_auglag$value 
    fitted.val_auglag <- exp(rbind(Xa1,Xaj1)[,-ncol(Xa)]%*%mle_auglag)
    Boundary0_auglag<-ifelse(any(fitted.val_auglag < 1e-5),1,0)
    Boundary1_auglag<-ifelse(any(fitted.val_auglag > 0.99999),1,0)
    SE_auglag<-c(NA,NA,NA,NA)
    try(SE_auglag<-sqrt(diag(solve(constr_optim_auglag$hessian))))
    conv_auglag<-constr_optim_auglag$convergence
    diffnum<-c(abs(mle_neldmead-mle_auglag),abs(loglik.max_neldmead-loglik.max_auglag))
  
    alld_neldmead <- abs(loglik.max_neldmead - max(loglik.max_neldmead, loglik.max_auglag))
    alld_auglag <- abs(loglik.max_auglag - max(loglik.max_neldmead, loglik.max_auglag))
  
    #print(r)
    outmatrix[r,]<-c(seedir,mle_neldmead,loglik.max_neldmead,Boundary0_neldmead,Boundary1_neldmead,
                     SE_constrOptim_neldmead,mle_auglag,loglik.max_auglag,Boundary0_auglag,
                     Boundary1_auglag, SE_auglag,diffnum,conv_neldmead,conv_auglag,
                     time_neldmead,time_auglag,alld_neldmead,alld_auglag)
  
  }

  currentDate <- format(Sys.Date(), "%m-%d-%y")
  rlog::log_trace(paste("writing output for r=", r))
  pfilename <- paste(basedir,'/output/P2-P0', nc, case,'-PPMcovars-', currentDate, '-T',n,'.out',sep='' )
  print(pfilename)
  write.table(outmatrix, file=pfilename,sep="\t",row.names=FALSE)
  
  rlog::log_trace(paste("generating metrics", pfilename))
  ### metrics

  #Relative % Bias_neldmead
  neldmead_bias_1 <- mean((outmatrix[,2]-alpha1)/(alpha1))*100
  neldmead_bias_2 <- mean((outmatrix[,3]-alpha2)/(alpha2))*100
  neldmead_bias_3 <- mean((outmatrix[,4]-beta1)/(beta1))*100
  neldmead_bias_4 <- mean((outmatrix[,5]-beta2)/(beta2))*100
  #cat('\n','neldmead_bias_1', neldmead_bias_1, 'neldmead_bias_2', neldmead_bias_2,
  #    'neldmead_bias_3', neldmead_bias_3, 'neldmead_bias_4', neldmead_bias_4)
  #Relative % Bias_auglag
  auglag_bias_1 <- mean((outmatrix[,13]-alpha1)/(alpha1))*100
  auglag_bias_2 <- mean((outmatrix[,14]-alpha2)/(alpha2))*100
  auglag_bias_3 <- mean((outmatrix[,15]-beta1)/(beta1))*100
  auglag_bias_4 <- mean((outmatrix[,16]-beta2)/(beta2))*100
  #cat('\n','auglag_bias_1', auglag_bias_1, 'auglag_bias_2', auglag_bias_2,
  #    'auglag_bias_3', auglag_bias_3, 'auglag_bias_4', auglag_bias_4)

  # ALLD

  alld_1 <- mean(outmatrix[,41])
  alld_2 <- mean(outmatrix[,42])
  #cat('\n','alld_1', alld_1, 'alld_2', alld_2)
  
  # RMSE
  #neldmead
  neldmead_rmse_1 <- sqrt(mean(((outmatrix[,2])-alpha1)^2))
  neldmead_rmse_2 <- sqrt(mean(((outmatrix[,3])-alpha2)^2))
  neldmead_rmse_3 <- sqrt(mean(((outmatrix[,4])-beta1)^2))
  neldmead_rmse_4 <- sqrt(mean(((outmatrix[,5])-beta2)^2))
  #cat('\n','neldmead_rmse_1', neldmead_rmse_1, 'neldmead_rmse_2', neldmead_rmse_2,
  #    'neldmead_rmse_3', neldmead_rmse_3, 'neldmead_rmse_4', neldmead_rmse_4)
  #auglag
  auglag_rmse_1 <- sqrt(mean(((outmatrix[,13])-alpha1)^2))
  auglag_rmse_2 <- sqrt(mean(((outmatrix[,14])-alpha2)^2))
  auglag_rmse_3 <- sqrt(mean(((outmatrix[,15])-beta1)^2))
  auglag_rmse_4 <- sqrt(mean(((outmatrix[,16])-beta2)^2))
  #cat('\n','auglag_rmse_1', auglag_rmse_1, 'auglag_rmse_2', auglag_rmse_2,
  #    'auglag_rmse_3', auglag_rmse_3, 'auglag_rmse_4', auglag_rmse_4)

  ## 

  #summary(outmatrix)

}